@@ -5,11 +5,12 @@
 #  SPDX-License-Identifier: LGPL-3.0-or-later
 #
 """
-Provides chromosome visitors to perform post processing.
+Provides chromosome visitors to perform post-processing.
 """
 from __future__ import annotations
 
 import logging
+from abc import ABC
 from typing import TYPE_CHECKING
 
 import pynguin.ga.chromosomevisitor as cv
@@ -38,7 +39,7 @@ class ExceptionTruncation(cv.ChromosomeVisitor):
 class TestCasePostProcessor(cv.ChromosomeVisitor):
     """Applies all given visitors to the visited test cases."""
 
-    def __init__(self, test_case_visitors: list[tcv.TestCaseVisitor]):
+    def __init__(self, test_case_visitors: list[ModificationAwareTestCaseVisitor]):
         self._test_case_visitors = test_case_visitors
 
     def visit_test_suite_chromosome(self, chromosome: tsc.TestSuiteChromosome) -> None:
@@ -48,15 +49,38 @@ class TestCasePostProcessor(cv.ChromosomeVisitor):
     def visit_test_case_chromosome(self, chromosome: tcc.TestCaseChromosome) -> None:
         for visitor in self._test_case_visitors:
             chromosome.test_case.accept(visitor)
+            if (last_exec := chromosome.get_last_execution_result()) is not None:
+                # We don't want to re-execute the test cases here, so we also remove
+                # information about the deleted statements from the execution result.
+                # TODO(fk) we could also re-execute, but with flakiness this could
+                #  cause inconsistent results
+                last_exec.delete_statement_data(visitor.deleted_statement_indexes)
+
+
+class ModificationAwareTestCaseVisitor(tcv.TestCaseVisitor, ABC):
+    """Visitor that keep information on modifications"""
+
+    def __init__(self):
+        self._deleted_statement_indexes: set[int] = set()
+
+    @property
+    def deleted_statement_indexes(self) -> set[int]:
+        """Provides a set of deleted statement indexes
+
+        Returns:
+            The deleted statement indexes
+        """
+        return self._deleted_statement_indexes
 
 
 # pylint:disable=too-few-public-methods
-class UnusedStatementsTestCaseVisitor(tcv.TestCaseVisitor):
+class UnusedStatementsTestCaseVisitor(ModificationAwareTestCaseVisitor):
     """Removes unused primitive and collection statements."""
 
     _logger = logging.getLogger(__name__)
 
     def visit_default_test_case(self, test_case) -> None:
+        self._deleted_statement_indexes.clear()
         primitive_remover = UnusedPrimitiveOrCollectionStatementVisitor()
         size_before = test_case.size()
         # Iterate over copy, to be able to modify original.
@@ -66,6 +90,9 @@ class UnusedStatementsTestCaseVisitor(tcv.TestCaseVisitor):
             "Removed %s unused primitives/collections from test case",
             size_before - test_case.size(),
         )
+        self._deleted_statement_indexes.update(
+            primitive_remover.deleted_statement_indexes
+        )
 
 
 class UnusedPrimitiveOrCollectionStatementVisitor(StatementVisitor):
@@ -74,11 +101,22 @@ class UnusedPrimitiveOrCollectionStatementVisitor(StatementVisitor):
 
     def __init__(self):
         self._used_references = set()
+        self._deleted_statement_indexes: set[int] = set()
+
+    @property
+    def deleted_statement_indexes(self) -> set[int]:
+        """Provides a set of deleted statement indexes
+
+        Returns:
+            The deleted statement indexes
+        """
+        return self._deleted_statement_indexes
 
     def _handle_collection_or_primitive(self, stmt) -> None:
         if stmt.ret_val in self._used_references:
             self._handle_remaining(stmt)
         else:
+            self._deleted_statement_indexes.add(stmt.get_position())
             stmt.test_case.remove_statement(stmt)
 
     def _handle_remaining(self, stmt) -> None:
