@@ -14,10 +14,8 @@ from typing import TYPE_CHECKING
 
 from bytecode import BasicBlock, Bytecode, Compare, ControlFlowGraph, Instr
 
-import pynguin.utils.opcodes as op
-from pynguin.analyses.controlflow.cfg import CFG
-from pynguin.analyses.controlflow.controldependencegraph import ControlDependenceGraph
-from pynguin.analyses.seeding.constantseeding import DynamicConstantSeeding
+from pynguin.analyses.constants import DynamicConstantProvider
+from pynguin.analyses.controlflow import CFG, ControlDependenceGraph
 from pynguin.testcase.execution import (
     CodeObjectMetaData,
     ExecutionTracer,
@@ -25,11 +23,16 @@ from pynguin.testcase.execution import (
 )
 
 if TYPE_CHECKING:
-    from pynguin.analyses.controlflow.programgraph import ProgramGraphNode
+    from pynguin.analyses.controlflow import ProgramGraphNode
 
 CODE_OBJECT_ID_KEY = "code_object_id"
 
 
+class ArtificialInstr(Instr):
+    """Marker subclass to distinguish between original instructions
+    and instructions that were inserted by the instrumentation."""
+
+
 # pylint:disable=too-few-public-methods
 class InstrumentationAdapter:
     """Abstract base class for bytecode instrumentation adapters.
@@ -51,11 +54,11 @@ class InstrumentationAdapter:
     # TODO(fk) make this more fine grained? e.g. visit_line, visit_compare etc.
     #  Or use sub visitors?
 
-    def visit_entry_node(self, block: BasicBlock, code_object_id: int) -> None:
+    def visit_entry_node(self, basic_block: BasicBlock, code_object_id: int) -> None:
         """Called when we visit the entry node of a code object.
 
         Args:
-            block: The basic block of the entry node.
+            basic_block: The basic block of the entry node.
             code_object_id: The code object id of the containing code object.
         """
 
@@ -96,7 +99,7 @@ class InstrumentationAdapter:
         current: BasicBlock = first
         nodes: list[BasicBlock] = []
         # Can be any instruction, as it is discarded anyway.
-        dummy_instruction = Instr("POP_TOP")
+        dummy_instruction = ArtificialInstr("POP_TOP")
         for _ in range(amount):
             # Insert dummy instruction, which we can use to split off another block
             current.insert(0, dummy_instruction)
@@ -111,6 +114,26 @@ class InstrumentationAdapter:
             node.clear()
         return tuple(nodes)
 
+    @staticmethod
+    def _map_instr_positions(basic_block: BasicBlock) -> dict[int, int]:
+        """Other instrumentations may add artificial instructions.
+        Create a mapping that maps original locations to their locations.
+
+        Args:
+            basic_block: The block that should be mapped.
+
+        Returns:
+            A mapping from original instructions positions to their new positions.
+        """
+        orig_idx = 0
+        mapping = {}
+        for idx, instr in enumerate(basic_block):
+            if isinstance(instr, ArtificialInstr):
+                continue
+            mapping[orig_idx] = idx
+            orig_idx += 1
+        return mapping
+
 
 class InstrumentationTransformer:
     """Applies a given list of instrumentation adapters to code objects.
@@ -235,14 +258,7 @@ class BranchCoverageInstrumentation(InstrumentationAdapter):
     """Instruments code objects to enable tracking branch distances and thus
     branch coverage."""
 
-    # As of CPython 3.8, there are a few compare ops for which we can't really
-    # compute a sensible branch distance. So for now, we just ignore those
-    # comparisons and just track their boolean value.
-    # As of CPython 3.9, this is no longer a compare op but instead
-    # a JUMP_IF_NOT_EXC_MATCH, which we also handle as boolean based jump.
-    _IGNORED_COMPARE_OPS: set[Compare] = {Compare.EXC_MATCH}
-
-    # Conditional jump operations are the last operation within a basic block
+    # Jump operations are the last operation within a basic block
     _JUMP_OP_POS = -1
 
     # If a conditional jump is based on a comparison, it has to be the second-to-last
@@ -274,22 +290,28 @@ class BranchCoverageInstrumentation(InstrumentationAdapter):
 
         assert len(basic_block) > 0, "Empty basic block in CFG."
         maybe_jump: Instr = basic_block[self._JUMP_OP_POS]
-        maybe_compare: Instr | None = (
-            basic_block[self._COMPARE_OP_POS] if len(basic_block) > 1 else None
+        orig_instructions_positions = InstrumentationAdapter._map_instr_positions(
+            basic_block
+        )
+        maybe_compare_idx: int | None = orig_instructions_positions.get(
+            len(orig_instructions_positions) + self._COMPARE_OP_POS
         )
         if isinstance(maybe_jump, Instr):
             predicate_id: int | None = None
             if maybe_jump.name == "FOR_ITER":
                 predicate_id = self._instrument_for_loop(
-                    cfg, node, basic_block, code_object_id
+                    cfg=cfg,
+                    node=node,
+                    basic_block=basic_block,
+                    code_object_id=code_object_id,
                 )
             elif maybe_jump.is_cond_jump():
                 predicate_id = self._instrument_cond_jump(
-                    code_object_id,
-                    maybe_compare,
-                    maybe_jump,
-                    basic_block,
-                    node,
+                    code_object_id=code_object_id,
+                    maybe_compare_idx=maybe_compare_idx,
+                    jump=maybe_jump,
+                    block=basic_block,
+                    node=node,
                 )
             if predicate_id is not None:
                 node.predicate_id = predicate_id
@@ -297,7 +319,7 @@ class BranchCoverageInstrumentation(InstrumentationAdapter):
     def _instrument_cond_jump(
         self,
         code_object_id: int,
-        maybe_compare: Instr | None,
+        maybe_compare_idx: int | None,
         jump: Instr,
         block: BasicBlock,
         node: ProgramGraphNode,
@@ -311,7 +333,7 @@ class BranchCoverageInstrumentation(InstrumentationAdapter):
 
         Args:
             code_object_id: The id of the containing Code Object.
-            maybe_compare: The comparison operation, if any.
+            maybe_compare_idx: The index of the comparison operation, if any.
             jump: The jump operation.
             block: The containing basic block.
             node: The associated node from the CFG.
@@ -319,28 +341,26 @@ class BranchCoverageInstrumentation(InstrumentationAdapter):
         Returns:
             The id that was assigned to the predicate.
         """
+        maybe_compare = block[maybe_compare_idx]
         if (
             maybe_compare is not None
             and isinstance(maybe_compare, Instr)
-            and (
-                (
-                    maybe_compare.name == "COMPARE_OP"
-                    and maybe_compare.arg not in self._IGNORED_COMPARE_OPS
-                )
-                or maybe_compare.name in ("IS_OP", "CONTAINS_OP")
-            )
+            and maybe_compare.name in ("COMPARE_OP", "IS_OP", "CONTAINS_OP")
         ):
+            assert maybe_compare_idx is not None
             return self._instrument_compare_based_conditional_jump(
-                block, code_object_id, node
+                block=block,
+                code_object_id=code_object_id,
+                compare_idx=maybe_compare_idx,
+                node=node,
             )
-        # Up to 3.9, there was COMPARE_OP EXC_MATCH which was handled below
-        # Beginning with 3.9, there is a combined compare+jump op, which is handled
-        # here.
         if jump.name == "JUMP_IF_NOT_EXC_MATCH":
             return self._instrument_exception_based_conditional_jump(
-                block, code_object_id, node
+                basic_block=block, code_object_id=code_object_id, node=node
             )
-        return self._instrument_bool_based_conditional_jump(block, code_object_id, node)
+        return self._instrument_bool_based_conditional_jump(
+            block=block, code_object_id=code_object_id, node=node
+        )
 
     def _instrument_bool_based_conditional_jump(
         self, block: BasicBlock, code_object_id: int, node: ProgramGraphNode
@@ -366,23 +386,27 @@ class BranchCoverageInstrumentation(InstrumentationAdapter):
         # We duplicate the value on top of the stack and report
         # it to the tracer.
         block[self._JUMP_OP_POS : self._JUMP_OP_POS] = [
-            Instr("DUP_TOP", lineno=lineno),
-            Instr("LOAD_CONST", self._tracer, lineno=lineno),
-            Instr(
+            ArtificialInstr("DUP_TOP", lineno=lineno),
+            ArtificialInstr("LOAD_CONST", self._tracer, lineno=lineno),
+            ArtificialInstr(
                 "LOAD_METHOD",
                 ExecutionTracer.executed_bool_predicate.__name__,
                 lineno=lineno,
             ),
-            Instr("ROT_THREE", lineno=lineno),
-            Instr("ROT_THREE", lineno=lineno),
-            Instr("LOAD_CONST", predicate_id, lineno=lineno),
-            Instr("CALL_METHOD", 2, lineno=lineno),
-            Instr("POP_TOP", lineno=lineno),
+            ArtificialInstr("ROT_THREE", lineno=lineno),
+            ArtificialInstr("ROT_THREE", lineno=lineno),
+            ArtificialInstr("LOAD_CONST", predicate_id, lineno=lineno),
+            ArtificialInstr("CALL_METHOD", 2, lineno=lineno),
+            ArtificialInstr("POP_TOP", lineno=lineno),
         ]
         return predicate_id
 
     def _instrument_compare_based_conditional_jump(
-        self, block: BasicBlock, code_object_id: int, node: ProgramGraphNode
+        self,
+        block: BasicBlock,
+        compare_idx: int,
+        code_object_id: int,
+        node: ProgramGraphNode,
     ) -> int:
         """Instrument compare-based conditional jumps.
 
@@ -391,6 +415,7 @@ class BranchCoverageInstrumentation(InstrumentationAdapter):
 
         Args:
             block: The containing basic block.
+            compare_idx: The index of the comparison index
             code_object_id: The id of the containing Code Object.
             node: The associated node from the CFG.
 
@@ -404,40 +429,43 @@ class BranchCoverageInstrumentation(InstrumentationAdapter):
         predicate_id = self._tracer.register_predicate(
             PredicateMetaData(line_no=lineno, code_object_id=code_object_id, node=node)
         )
-        operation = block[self._COMPARE_OP_POS]
+        operation = block[compare_idx]
 
-        if operation.name == "COMPARE_OP":
-            compare = operation.arg
-        elif operation.name == "IS_OP":
-            # Beginning with 3.9, there are separate OPs for various comparisons.
-            compare = Compare.IS_NOT if operation.arg else Compare.IS
-        elif operation.name == "CONTAINS_OP":
-            compare = Compare.NOT_IN if operation.arg else Compare.IN
-        else:
-            raise RuntimeError(f"Unknown comparison OP {operation}")
+        match operation.name:
+            case "COMPARE_OP":
+                compare = operation.arg
+            case "IS_OP":
+                # Beginning with 3.9, there are separate OPs for various comparisons.
+                # Map them back to the old operations, so we can use the enum from the
+                # bytecode library.
+                compare = Compare.IS_NOT if operation.arg else Compare.IS
+            case "CONTAINS_OP":
+                compare = Compare.NOT_IN if operation.arg else Compare.IN
+            case _:
+                raise RuntimeError(f"Unknown comparison OP {operation}")
 
         # Insert instructions right before the comparison.
         # We duplicate the values on top of the stack and report
         # them to the tracer.
-        block[self._COMPARE_OP_POS : self._COMPARE_OP_POS] = [
-            Instr("DUP_TOP_TWO", lineno=lineno),
-            Instr("LOAD_CONST", self._tracer, lineno=lineno),
-            Instr(
+        block[compare_idx:compare_idx] = [
+            ArtificialInstr("DUP_TOP_TWO", lineno=lineno),
+            ArtificialInstr("LOAD_CONST", self._tracer, lineno=lineno),
+            ArtificialInstr(
                 "LOAD_METHOD",
                 ExecutionTracer.executed_compare_predicate.__name__,
                 lineno=lineno,
             ),
-            Instr("ROT_FOUR", lineno=lineno),
-            Instr("ROT_FOUR", lineno=lineno),
-            Instr("LOAD_CONST", predicate_id, lineno=lineno),
-            Instr("LOAD_CONST", compare, lineno=lineno),
-            Instr("CALL_METHOD", 4, lineno=lineno),
-            Instr("POP_TOP", lineno=lineno),
+            ArtificialInstr("ROT_FOUR", lineno=lineno),
+            ArtificialInstr("ROT_FOUR", lineno=lineno),
+            ArtificialInstr("LOAD_CONST", predicate_id, lineno=lineno),
+            ArtificialInstr("LOAD_CONST", compare, lineno=lineno),
+            ArtificialInstr("CALL_METHOD", 4, lineno=lineno),
+            ArtificialInstr("POP_TOP", lineno=lineno),
         ]
         return predicate_id
 
     def _instrument_exception_based_conditional_jump(
-        self, block: BasicBlock, code_object_id: int, node: ProgramGraphNode
+        self, basic_block: BasicBlock, code_object_id: int, node: ProgramGraphNode
     ) -> int:
         """Instrument exception-based conditional jumps.
 
@@ -445,58 +473,59 @@ class BranchCoverageInstrumentation(InstrumentationAdapter):
         in the following exception matching case.
 
         Args:
-            block: The containing basic block.
+            basic_block: The containing basic block.
             code_object_id: The id of the containing Code Object.
             node: The associated node from the CFG.
 
         Returns:
             The id assigned to the predicate.
         """
-        lineno = block[self._JUMP_OP_POS].lineno
+        lineno = basic_block[self._JUMP_OP_POS].lineno
         predicate_id = self._tracer.register_predicate(
             PredicateMetaData(line_no=lineno, code_object_id=code_object_id, node=node)
         )
         # Insert instructions right before the conditional jump.
         # We duplicate the values on top of the stack and report
         # them to the tracer.
-        block[self._JUMP_OP_POS : self._JUMP_OP_POS] = [
-            Instr("DUP_TOP_TWO", lineno=lineno),
-            Instr("LOAD_CONST", self._tracer, lineno=lineno),
-            Instr(
+        basic_block[self._JUMP_OP_POS : self._JUMP_OP_POS] = [
+            ArtificialInstr("DUP_TOP_TWO", lineno=lineno),
+            ArtificialInstr("LOAD_CONST", self._tracer, lineno=lineno),
+            ArtificialInstr(
                 "LOAD_METHOD",
                 ExecutionTracer.executed_exception_match.__name__,
                 lineno=lineno,
             ),
-            Instr("ROT_FOUR", lineno=lineno),
-            Instr("ROT_FOUR", lineno=lineno),
-            Instr("LOAD_CONST", predicate_id, lineno=lineno),
-            Instr("CALL_METHOD", 3, lineno=lineno),
-            Instr("POP_TOP", lineno=lineno),
+            ArtificialInstr("ROT_FOUR", lineno=lineno),
+            ArtificialInstr("ROT_FOUR", lineno=lineno),
+            ArtificialInstr("LOAD_CONST", predicate_id, lineno=lineno),
+            ArtificialInstr("CALL_METHOD", 3, lineno=lineno),
+            ArtificialInstr("POP_TOP", lineno=lineno),
         ]
         return predicate_id
 
-    def visit_entry_node(self, block: BasicBlock, code_object_id: int) -> None:
+    def visit_entry_node(self, basic_block: BasicBlock, code_object_id: int) -> None:
         """Add instructions at the beginning of the given basic block which inform
         the tracer, that the code object with the given id has been entered.
 
         Args:
-            block: The entry basic block of a code object, i.e. the first basic block.
+            basic_block: The entry basic block of a code object, i.e. the first basic
+                block.
             code_object_id: The id that the tracer has assigned to the code object
                 which contains the given basic block.
         """
         # Use line number of first instruction
-        lineno = block[0].lineno
+        lineno = basic_block[0].lineno
         # Insert instructions at the beginning.
-        block[0:0] = [
-            Instr("LOAD_CONST", self._tracer, lineno=lineno),
-            Instr(
+        basic_block[0:0] = [
+            ArtificialInstr("LOAD_CONST", self._tracer, lineno=lineno),
+            ArtificialInstr(
                 "LOAD_METHOD",
                 ExecutionTracer.executed_code_object.__name__,
                 lineno=lineno,
             ),
-            Instr("LOAD_CONST", code_object_id, lineno=lineno),
-            Instr("CALL_METHOD", 1, lineno=lineno),
-            Instr("POP_TOP", lineno=lineno),
+            ArtificialInstr("LOAD_CONST", code_object_id, lineno=lineno),
+            ArtificialInstr("CALL_METHOD", 1, lineno=lineno),
+            ArtificialInstr("POP_TOP", lineno=lineno),
         ]
 
     def _instrument_for_loop(
@@ -552,37 +581,39 @@ class BranchCoverageInstrumentation(InstrumentationAdapter):
         entered, not_entered = self._create_consecutive_blocks(
             cfg.bytecode_cfg(), basic_block, 2
         )
+        # TODO(fk) for_instr is not artificial but we changed it
+        #  How to deal with this?
         for_instr.arg = not_entered
 
         entered.extend(
             [
-                Instr("LOAD_CONST", self._tracer, lineno=lineno),
-                Instr(
+                ArtificialInstr("LOAD_CONST", self._tracer, lineno=lineno),
+                ArtificialInstr(
                     "LOAD_METHOD",
                     ExecutionTracer.executed_bool_predicate.__name__,
                     lineno=lineno,
                 ),
-                Instr("LOAD_CONST", True, lineno=lineno),
-                Instr("LOAD_CONST", predicate_id, lineno=lineno),
-                Instr("CALL_METHOD", 2, lineno=lineno),
-                Instr("POP_TOP", lineno=lineno),
-                Instr("JUMP_ABSOLUTE", for_loop_body, lineno=lineno),
+                ArtificialInstr("LOAD_CONST", True, lineno=lineno),
+                ArtificialInstr("LOAD_CONST", predicate_id, lineno=lineno),
+                ArtificialInstr("CALL_METHOD", 2, lineno=lineno),
+                ArtificialInstr("POP_TOP", lineno=lineno),
+                ArtificialInstr("JUMP_ABSOLUTE", for_loop_body, lineno=lineno),
             ]
         )
 
         not_entered.extend(
             [
-                Instr("LOAD_CONST", self._tracer, lineno=lineno),
-                Instr(
+                ArtificialInstr("LOAD_CONST", self._tracer, lineno=lineno),
+                ArtificialInstr(
                     "LOAD_METHOD",
                     ExecutionTracer.executed_bool_predicate.__name__,
                     lineno=lineno,
                 ),
-                Instr("LOAD_CONST", False, lineno=lineno),
-                Instr("LOAD_CONST", predicate_id, lineno=lineno),
-                Instr("CALL_METHOD", 2, lineno=lineno),
-                Instr("POP_TOP", lineno=lineno),
-                Instr("JUMP_ABSOLUTE", for_loop_exit, lineno=lineno),
+                ArtificialInstr("LOAD_CONST", False, lineno=lineno),
+                ArtificialInstr("LOAD_CONST", predicate_id, lineno=lineno),
+                ArtificialInstr("CALL_METHOD", 2, lineno=lineno),
+                ArtificialInstr("POP_TOP", lineno=lineno),
+                ArtificialInstr("JUMP_ABSOLUTE", for_loop_exit, lineno=lineno),
             ]
         )
 
@@ -606,22 +637,19 @@ class LineCoverageInstrumentation(InstrumentationAdapter):
         node: ProgramGraphNode,
         basic_block: BasicBlock,
     ) -> None:
-        if not is_return_none_basic_block(basic_block):
-            #  iterate over instructions after the fist one in BB,
-            #  put new instructions in the block for each line
-            file_name = cfg.bytecode_cfg().filename
-            lineno = -1
-            instr_index = 0
-            while instr_index < len(basic_block):
-                if basic_block[instr_index].lineno != lineno:
-                    lineno = basic_block[instr_index].lineno
-                    line_id = self._tracer.register_line(
-                        code_object_id, file_name, lineno
-                    )
-                    instr_index += (  # increment by the amount of instructions inserted
-                        self.instrument_line(basic_block, instr_index, line_id, lineno)
-                    )
-                instr_index += 1
+        #  iterate over instructions after the fist one in BB,
+        #  put new instructions in the block for each line
+        file_name = cfg.bytecode_cfg().filename
+        lineno = None
+        instr_index = 0
+        while instr_index < len(basic_block):
+            if basic_block[instr_index].lineno != lineno:
+                lineno = basic_block[instr_index].lineno
+                line_id = self._tracer.register_line(code_object_id, file_name, lineno)
+                instr_index += (  # increment by the amount of instructions inserted
+                    self.instrument_line(basic_block, instr_index, line_id, lineno)
+                )
+            instr_index += 1
 
     def instrument_line(
         self, block: BasicBlock, instr_index: int, line_id: int, lineno: int
@@ -640,15 +668,15 @@ class LineCoverageInstrumentation(InstrumentationAdapter):
             The number of instructions inserted into the block
         """
         inserted_instructions = [
-            Instr("LOAD_CONST", self._tracer, lineno=lineno),
-            Instr(
+            ArtificialInstr("LOAD_CONST", self._tracer, lineno=lineno),
+            ArtificialInstr(
                 "LOAD_METHOD",
                 self._tracer.track_line_visit.__name__,
                 lineno=lineno,
             ),
-            Instr("LOAD_CONST", line_id, lineno=lineno),
-            Instr("CALL_METHOD", 1, lineno=lineno),
-            Instr("POP_TOP", lineno=lineno),
+            ArtificialInstr("LOAD_CONST", line_id, lineno=lineno),
+            ArtificialInstr("CALL_METHOD", 1, lineno=lineno),
+            ArtificialInstr("POP_TOP", lineno=lineno),
         ]
         # Insert instructions at the beginning.
         block[instr_index:instr_index] = inserted_instructions
@@ -685,27 +713,10 @@ class DynamicSeedingInstrumentation(InstrumentationAdapter):
     # operation.
     _STRING_FUNC_POS_WITH_ARG = -4
 
-    # A list containing the names of all string functions which are instrumented.
-    _STRING_FUNCTION_NAMES = [
-        "startswith",
-        "endswith",
-        "isalnum",
-        "isalpha",
-        "isdecimal",
-        "isdigit",
-        "isidentifier",
-        "islower",
-        "isnumeric",
-        "isprintable",
-        "isspace",
-        "istitle",
-        "isupper",
-    ]
-
     _logger = logging.getLogger(__name__)
 
-    def __init__(self, dynamic_constant_seeding: DynamicConstantSeeding):
-        self._dynamic_constant_seeding = dynamic_constant_seeding
+    def __init__(self, dynamic_constant_provider: DynamicConstantProvider):
+        self._dynamic_constant_provider = dynamic_constant_provider
 
     def visit_node(
         self,
@@ -731,13 +742,13 @@ class DynamicSeedingInstrumentation(InstrumentationAdapter):
         if (
             isinstance(maybe_string_func, Instr)
             and maybe_string_func.name == "LOAD_METHOD"
-            and maybe_string_func.arg in self._STRING_FUNCTION_NAMES
+            and maybe_string_func.arg in DynamicConstantProvider.STRING_FUNCTION_LOOKUP
         ):
             self._instrument_string_func(basic_block, maybe_string_func.arg)
         if (
             isinstance(maybe_string_func_with_arg, Instr)
             and maybe_string_func_with_arg.name == "LOAD_METHOD"
-            and maybe_string_func_with_arg.arg in self._STRING_FUNCTION_NAMES
+            and maybe_string_func_with_arg.arg in {"startswith", "endswith"}
         ):
             self._instrument_string_func(basic_block, maybe_string_func_with_arg.arg)
 
@@ -752,19 +763,21 @@ class DynamicSeedingInstrumentation(InstrumentationAdapter):
         insert_pos = self._STRING_FUNC_POS_WITH_ARG + 2
         lineno = block[insert_pos].lineno
         block[insert_pos:insert_pos] = [
-            Instr("DUP_TOP_TWO", lineno=lineno),
-            Instr("ROT_TWO", lineno=lineno),
-            Instr("BINARY_ADD", lineno=lineno),
-            Instr("LOAD_CONST", self._dynamic_constant_seeding, lineno=lineno),
-            Instr(
+            ArtificialInstr("DUP_TOP_TWO", lineno=lineno),
+            ArtificialInstr("ROT_TWO", lineno=lineno),
+            ArtificialInstr("BINARY_ADD", lineno=lineno),
+            ArtificialInstr(
+                "LOAD_CONST", self._dynamic_constant_provider, lineno=lineno
+            ),
+            ArtificialInstr(
                 "LOAD_METHOD",
-                self._dynamic_constant_seeding.add_value.__name__,
+                DynamicConstantProvider.add_value.__name__,
                 lineno=lineno,
             ),
-            Instr("ROT_THREE", lineno=lineno),
-            Instr("ROT_THREE", lineno=lineno),
-            Instr("CALL_METHOD", 1, lineno=lineno),
-            Instr("POP_TOP", lineno=lineno),
+            ArtificialInstr("ROT_THREE", lineno=lineno),
+            ArtificialInstr("ROT_THREE", lineno=lineno),
+            ArtificialInstr("CALL_METHOD", 1, lineno=lineno),
+            ArtificialInstr("POP_TOP", lineno=lineno),
         ]
         self._logger.info("Instrumented startswith function")
 
@@ -779,18 +792,20 @@ class DynamicSeedingInstrumentation(InstrumentationAdapter):
         insert_pos = self._STRING_FUNC_POS_WITH_ARG + 2
         lineno = block[insert_pos].lineno
         block[insert_pos:insert_pos] = [
-            Instr("DUP_TOP_TWO", lineno=lineno),
-            Instr("BINARY_ADD", lineno=lineno),
-            Instr("LOAD_CONST", self._dynamic_constant_seeding, lineno=lineno),
-            Instr(
+            ArtificialInstr("DUP_TOP_TWO", lineno=lineno),
+            ArtificialInstr("BINARY_ADD", lineno=lineno),
+            ArtificialInstr(
+                "LOAD_CONST", self._dynamic_constant_provider, lineno=lineno
+            ),
+            ArtificialInstr(
                 "LOAD_METHOD",
-                DynamicConstantSeeding.add_value.__name__,
+                DynamicConstantProvider.add_value.__name__,
                 lineno=lineno,
             ),
-            Instr("ROT_THREE", lineno=lineno),
-            Instr("ROT_THREE", lineno=lineno),
-            Instr("CALL_METHOD", 1, lineno=lineno),
-            Instr("POP_TOP", lineno=lineno),
+            ArtificialInstr("ROT_THREE", lineno=lineno),
+            ArtificialInstr("ROT_THREE", lineno=lineno),
+            ArtificialInstr("CALL_METHOD", 1, lineno=lineno),
+            ArtificialInstr("POP_TOP", lineno=lineno),
         ]
         self._logger.info("Instrumented endswith function")
 
@@ -806,18 +821,20 @@ class DynamicSeedingInstrumentation(InstrumentationAdapter):
         insert_pos = self._STRING_FUNC_POS_WITH_ARG + 2
         lineno = block[insert_pos].lineno
         block[insert_pos:insert_pos] = [
-            Instr("DUP_TOP", lineno=lineno),
-            Instr("LOAD_CONST", self._dynamic_constant_seeding, lineno=lineno),
-            Instr(
+            ArtificialInstr("DUP_TOP", lineno=lineno),
+            ArtificialInstr(
+                "LOAD_CONST", self._dynamic_constant_provider, lineno=lineno
+            ),
+            ArtificialInstr(
                 "LOAD_METHOD",
-                DynamicConstantSeeding.add_value_for_strings.__name__,
+                DynamicConstantProvider.add_value_for_strings.__name__,
                 lineno=lineno,
             ),
-            Instr("ROT_THREE", lineno=lineno),
-            Instr("ROT_THREE", lineno=lineno),
-            Instr("LOAD_CONST", function_name, lineno=lineno),
-            Instr("CALL_METHOD", 2, lineno=lineno),
-            Instr("POP_TOP", lineno=lineno),
+            ArtificialInstr("ROT_THREE", lineno=lineno),
+            ArtificialInstr("ROT_THREE", lineno=lineno),
+            ArtificialInstr("LOAD_CONST", function_name, lineno=lineno),
+            ArtificialInstr("CALL_METHOD", 2, lineno=lineno),
+            ArtificialInstr("POP_TOP", lineno=lineno),
         ]
         self._logger.info("Instrumented string function")
 
@@ -845,44 +862,30 @@ class DynamicSeedingInstrumentation(InstrumentationAdapter):
         """
         lineno = block[self._COMPARE_OP_POS].lineno
         block[self._COMPARE_OP_POS : self._COMPARE_OP_POS] = [
-            Instr("DUP_TOP_TWO", lineno=lineno),
-            Instr("LOAD_CONST", self._dynamic_constant_seeding, lineno=lineno),
-            Instr(
+            ArtificialInstr("DUP_TOP_TWO", lineno=lineno),
+            ArtificialInstr(
+                "LOAD_CONST", self._dynamic_constant_provider, lineno=lineno
+            ),
+            ArtificialInstr(
                 "LOAD_METHOD",
-                DynamicConstantSeeding.add_value.__name__,
+                DynamicConstantProvider.add_value.__name__,
                 lineno=lineno,
             ),
-            Instr("ROT_THREE", lineno=lineno),
-            Instr("ROT_THREE", lineno=lineno),
-            Instr("CALL_METHOD", 1, lineno=lineno),
-            Instr("POP_TOP", lineno=lineno),
-            Instr("LOAD_CONST", self._dynamic_constant_seeding, lineno=lineno),
-            Instr(
+            ArtificialInstr("ROT_THREE", lineno=lineno),
+            ArtificialInstr("ROT_THREE", lineno=lineno),
+            ArtificialInstr("CALL_METHOD", 1, lineno=lineno),
+            ArtificialInstr("POP_TOP", lineno=lineno),
+            ArtificialInstr(
+                "LOAD_CONST", self._dynamic_constant_provider, lineno=lineno
+            ),
+            ArtificialInstr(
                 "LOAD_METHOD",
-                DynamicConstantSeeding.add_value.__name__,
+                DynamicConstantProvider.add_value.__name__,
                 lineno=lineno,
             ),
-            Instr("ROT_THREE", lineno=lineno),
-            Instr("ROT_THREE", lineno=lineno),
-            Instr("CALL_METHOD", 1, lineno=lineno),
-            Instr("POP_TOP", lineno=lineno),
+            ArtificialInstr("ROT_THREE", lineno=lineno),
+            ArtificialInstr("ROT_THREE", lineno=lineno),
+            ArtificialInstr("CALL_METHOD", 1, lineno=lineno),
+            ArtificialInstr("POP_TOP", lineno=lineno),
         ]
         self._logger.debug("Instrumented compare_op")
-
-
-def is_return_none_basic_block(basic_block: BasicBlock) -> bool:
-    """Checks if a node is a "return None" line.
-
-    Args:
-        basic_block: The basic block that needs to be checked
-
-    Returns:
-        True, if the node is a "return None" line, false otherwise.
-    """
-    # TODO(fk) there seem to be cases where this check is not sufficient.
-    # Not sure how to detect those.
-    return (
-        len(basic_block) == 2
-        and basic_block[0] == Instr("LOAD_CONST", None, lineno=basic_block[0].lineno)
-        and basic_block[1].opcode == op.RETURN_VALUE
-    )
